/* Copyright Faeem Ali 
 * email: kodgehopper@netscape.net
 *
 *----------------------------------------------------------------------
 * This file is part of quickrun 
 *
 *  quickrun is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  quickrun is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with quickrun in the file COPYING; if not, write to the Free Software
 *  Foundation, Inc.:
 *      59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *----------------------------------------------------------------------
 */

#include "runbar.h"
#include "complete.h"

/*
 * strictly a debugging function.
 * shows the value of list...i.e. the dropdown list in the combobox
 */
void show_list()
{
   GList *curr;
   
   printf("->start of list:\n");
   curr=list;
   while (curr != NULL)
     {
	char *p;
	p=(char *)(curr->data);
	printf("%s\n",p);
	curr=curr->next;
     }   
   printf("->end of list\n");
}//show_list

void show_help(char *name)
{
   printf("Usage: %s [-h] [-p position] [-c count] [-s] [-k]\n\n",name);
   printf("*-h\t shows this help message\n");
   printf("*-p\t position can be \"none\", \"mouse\" or \"center\".\n");
   printf("   \t This determines where the window will pop up.\n");
   printf("   \t By default, the window pops up under the mouse\n");
   printf("   \t pointer. \"none\" lets the window manager choose\n");
   printf("   \t the best position and \"center\" displays the window\n");
   printf("   \t in the center of the screen.\n");
   printf("   \t \"centre\" is the same as \"center\". It's there\n");
   printf("   \t since the makers of English screwed up spelling a bit.\n");
   printf("*-c\t The number of commands to store in the history list.\n");
   printf("   \t Defaults to 10\n");
   printf("*-s\t Allow only a single instance of the program to run.\n");
   printf("   \t If another copy of quickrun is already running, then\n");
   printf("   \t the newly run process will quietly exit.\n");
   printf("*-k\t Keep the window open after executing a command. Useful\n");
   printf("   \t if you want to execute more than one command before\n");
   printf("   \t exiting. The window must be manually closed with the\n");
   printf("   \t mouse or Alt+q\n");
   
   exit(0);
} //show_help

/* callback for delete window */
void delete_window(GtkWidget *window, gpointer data)
{
   /* unlock lock file if -s specified */
   flock(fd, LOCK_UN);
   close(fd);
   
   /* and say bye bye */
   gtk_main_quit();
} //delete_window


/*
 * get data from combo box, add to history list and execute
 */
void combobox_enter(GtkWidget *widget, GdkEvent *event, gpointer data)
{
   FILE *fp;
   char cmd[1024]="";
   GList *tmp;
   guint list_length;  
   char *t;
   int prepend=1;
   
   /* create command */   
   strcpy(cmd,gtk_entry_get_text(GTK_ENTRY(widget)));
      
   /* put & at end if necessary and execute */
   if ((t=strchr(cmd,'&'))==NULL)
     {
	strcat(cmd,"&");
     }
   
#ifdef _DEBUG_
   printf("combobox_enter: command to execute: \"%s\"\n",cmd);
#endif


   if (strlen(cmd) > 1) //command + &
     system(cmd);
   
   if ((t=strchr(cmd,'&'))!=NULL)
     {
	*t='\0';
     }
   
   if (strlen(cmd) >= 1) //length of whatever the command was + \n
     {
	/* prepend command to history list if it's not in the list already */
	tmp=list;
	while (tmp != NULL)
	  {
	     char *p;
	     
	     p=(char *)(tmp->data);
	     if (!strcmp(p,cmd))
	       prepend=0;
	     
	     tmp=tmp->next;
	  }//while
	
	if (prepend) 
	  list=g_list_prepend(list,strdup(cmd));
	
	/* 
	 * truncate list if necessary.
	 * this is so we dont store more than the user specified
	 * number of elements.
	 */
	list_length=g_list_length(list);
	if (list_length > max_num_history)
	  {
	     GList *t_list, *t_list2;
	     
	     tmp=g_list_nth(list,max_num_history);
	     
	     t_list=tmp->next;
	     tmp->next=NULL;
	     
	     /* free rest of list */
	     t_list2=t_list;
	     while (t_list2 != NULL)
	       {
		  free(t_list2->data);
		  t_list2=t_list2->next;
	       }	     
	     g_list_free(t_list);
	  }
	
	/* write list to history file */
	if ((fp=fopen(filename,"w"))==NULL)
	  {
	     perror("error opening history file for writing");
	     exit(-1);
	  }
	
	tmp=list;
	while (tmp != NULL)
	  {
	     char *p;
	     
	     p=((char *)(tmp->data));
	     fprintf(fp,"%s\n",p);
	     tmp=tmp->next;
	  }       	
	fclose(fp);	
     }
   
   /* keep==1 means leave the window open after we execute a command */
   if (keep) 
     {
	/* refresh list of entries and clear text box */
	gtk_combo_set_popdown_strings(GTK_COMBO(combobox), list);
	gtk_entry_set_text(GTK_ENTRY(widget), "");
     }
   else
     {	
	gtk_main_quit();
     }   
}//combobox_enter


/*
 * reads the history list from ~/.quickrun_history and stores in "list"
 */
GList *get_history_list()
{
   FILE *fp=NULL;
   int homedir_len=0;
   char cmd[256];
   GList *list=NULL;
   
   homedir_len=strlen(homedir)+strlen("/.quickrun_history");
   if ((filename=(char *)malloc(homedir_len+1))==NULL)
     {
	perror("allocting memory for history file");
	exit(-1);
     }
   sprintf(filename,"%s/.quickrun_history",homedir);
   
   /*
    * open history file and read data. create one if it doesnt exist
    */
   if ((fp=fopen(filename,"r"))==NULL)
     {
	/* create new file */
	if ((fp=fopen(filename,"w"))==NULL)
	  {
	     perror("error creating history file");
	     exit(-1);
	  }
	
	fclose(fp);
	return NULL;
     }

   fgets(cmd,256,fp);
   while (!feof(fp))
     {
	char *p;
	
	/* remove newline */
	if ((p=strchr(cmd,'\n'))!=NULL)
	  {
	     *p='\0';
	  }	  
	list=g_list_append(list,strdup(cmd));
	fgets(cmd,256,fp);
     }

   fclose(fp);
        
   return list;
}//get_history_list

/* 
 * if -s is set, then the program should exit if another copy
 * is already running. returns 0 if program should exit, 
 * 1 if program should continue. -1 = error
 * 
 * A lock file is created here, and held for the duration of the
 * program runtime. the file is released when the program exits.
 */
int check_single()
{
   char *lock_file;
   int lock_file_len;
   
   if (!single)
     return 1;
   
   lock_file_len=strlen(homedir)+strlen("/.quickrun_lock");
   if ((lock_file=(char *)malloc(lock_file_len+1))==NULL)
     {
	perror("error allocating memory for filename");
	return -1;
     }
   
   /* make filename */
   sprintf(lock_file,"%s/.quickrun_lock",homedir);
   
   if ((fd=open(lock_file,O_CREAT))==-1)
     {
	perror("error creating lock file");
     }
   
   /* lock the file */
   if (flock(fd,LOCK_EX|LOCK_NB) == -1)
     {
	return 0; //return (and exit) if we cant get a lock on the file
     }
      
   /* clean up */
   free(lock_file);
   return 1;
}//check_single

/* shows a blank list. useful if there's nothing useful to show eg.
 * there's no match
 */
void show_blank_list()
{
   gtk_clist_clear(GTK_CLIST(clist));
}//show_blank_list

/*
 * determines what happens when a row in the popup list is
 * selected. a normal text complete takes place
 */
void popup_row_selected(GtkCList *clist, gint row, gint col, 
			GdkEventButton *event, gpointer data)
{
   GtkEntry *textbox=NULL;
   char *t_box=NULL;
   int t_box_len=0;
   int cursor_pos=0;
   char *buffer=NULL;
   char *new_text=NULL;
   
   /* retrieve the text from the text box, the cursor position and
    * selected text from the popup    
    */
   textbox=GTK_ENTRY(GTK_COMBO(combobox)->entry);
   t_box=gtk_entry_get_text(textbox);
   t_box_len=strlen(t_box);
   
   cursor_pos=gtk_editable_get_position(GTK_EDITABLE(textbox));
   
   gtk_clist_get_text(clist, row, col, &buffer);   
    
   /* retrieve the new, modified string and update textbox. also set cursor */
   new_text=get_known_completion(t_box, cursor_pos, buffer);
   cursor_pos += (strlen(new_text) - t_box_len);
   if (cursor_pos < 0) cursor_pos=0;
   
   if (new_text != NULL)
     {
	gtk_entry_set_text(textbox, new_text);	
	gtk_entry_set_position(textbox, cursor_pos);
	if (f_match.show_list) 
	  show_completions_list();
	else
	  show_blank_list();
	free(new_text);
     }
}//popup_row_selected


/* sorts out the data from 2 GList nodes in ascending order */
gint sort_nodes_ascending(gconstpointer a, gconstpointer b)
{
   char *s1, *s2;
      
   s1=(char *)a;
   s2=(char *)b;
     
   return strcasecmp(s1,s2);
}//sort_nodes_ascending

/*
 * extend the list in the main window to include the list of
 * completions stored if f_match. The first time this is called,
 * the list will have to be created. after that, the list must be
 * updated as necessary.
 * 
 * the main window will have to be resized to accommodate the change
 */
void show_completions_list()
{
   GList *curr=NULL;
   
   if (!window_extended)
     {	
	window_extended=1;
	clist=gtk_clist_new(1);
	gtk_signal_connect(GTK_OBJECT(clist),"select_row",GTK_SIGNAL_FUNC(popup_row_selected),NULL);
	/* sort out entries in list automatically */

	/* these aren't used cos they screw up the focus after sorting */
	//gtk_clist_set_sort_column(GTK_CLIST(clist), 0);
	//gtk_clist_set_auto_sort(GTK_CLIST(clist), TRUE);
	
	
	cscrollwindow=gtk_scrolled_window_new(NULL,NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(cscrollwindow),
				       GTK_POLICY_AUTOMATIC,
				       GTK_POLICY_AUTOMATIC);
	/* we're really interested in the height here, since the
	 * hbox makes the widget take up all available space
	 */
	gtk_widget_set_usize(cscrollwindow,50,100);
	
	gtk_container_add(GTK_CONTAINER(cscrollwindow), clist);
	gtk_widget_show(clist);
	
	gtk_container_add(GTK_CONTAINER(completionhbox), cscrollwindow);
	gtk_widget_show(cscrollwindow);
     }   

   /* clean out the old items in the clist */
   gtk_clist_clear(GTK_CLIST(clist));
  
   /* sort our f_match in ascending order 
    * we do this here instead of have the list do autosorting
    * since the list moves the focus away from the first element,
    * which complicates things a bit when using the keyboard
    */
   f_match.files=g_list_sort(f_match.files, sort_nodes_ascending);
   
   /* add new text to the list */
   curr=f_match.files;
   while (curr != NULL)
     {
	char *p;
	
	p=(char *)(curr->data);
	gtk_clist_append(GTK_CLIST(clist), &p);
	curr=curr->next;
     }//while
   
   /* move the list up to the first element */
   gtk_clist_moveto(GTK_CLIST(clist), 0, 0, 0, 0);
}//show_completions_list


/* activates filename completion */
gboolean text_complete(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
   char *new_text=NULL;
   
   /* when Esc is pressed it signals autocomplete time.
    * basically, all we do here is get the text and the position
    * of the cursor and hand it over to other functions to do
    * the hard work of completion. then we come back and display
    * the results in a popup window
    * 
    */
   if (event->keyval==GDK_Escape)
     {
	GtkEntry *textbox=NULL;
	gint cursor_pos=0;
	char *buffer=NULL;
	int buffer_len=0;
	int new_cursor_pos=0; //for re-positioning the cursor
	
	textbox=GTK_ENTRY(GTK_COMBO(combobox)->entry);
	cursor_pos=gtk_editable_get_position(GTK_EDITABLE(textbox));
	buffer=gtk_entry_get_text(textbox);
	buffer_len=strlen(buffer);
	
#ifdef _DEBUG_
	printf("cursor position: %d\n",cursor_pos);
	printf("text from textbox: \"%s\"\n",buffer);
#endif
	
	/* get text completions. these will be stored in
	 * the relevant global variables */
	new_text=get_completions(buffer,cursor_pos);
	
	
	/* memory is allocated for new_text. so it must be free()'ed.
	 * memory is allocated in one of the functions get_completions()
	 * uses
	 */

	if (new_text != NULL) //NULL==error so dont do anything
	  {
	     /* determine where to reposition the cursor */
	     new_cursor_pos=cursor_pos + (strlen(new_text) - buffer_len);
	     if (new_cursor_pos < 0) new_cursor_pos=0; //error condition

	     gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(combobox)->entry), new_text);
	     gtk_entry_set_position(GTK_ENTRY(GTK_COMBO(combobox)->entry), new_cursor_pos);
	     if (f_match.show_list) 
	       show_completions_list();
	     else
	       show_blank_list();
	     
	     free(new_text);
	  }	
     } //if escape pressed
   
   return TRUE;
}//text_complete

/* basic initializations, like getting home directories etc etc */
void initialize()
{
   /* get pointer to path environment variable */
   if ((path_string_ptr=getenv("PATH"))==NULL)
     {
	fprintf(stderr,"error determining path from $PATH\n");
	exit(-1);       
     }
   
   /* always work with a duplicate since we dont wanna change the
    * environment setting */
   path_string=strdup(path_string_ptr);
   
#ifdef _DEBUG_   
   printf("->path: %s\n",path_string);
#endif
   
   /* get home dir */
   if ((homedir_ptr=getenv("HOME"))==NULL)
     {
	fprintf(stderr,"error getting home directory. setting to \"/\"\n");
	homedir=strdup("/");
     }
   else
     {
	homedir=strdup(homedir_ptr);	
     }
   
   
#ifdef _DEBUG_
   printf("->home directory: %s\n",homedir);
#endif
}//initialize


/*
 * process command line arguments 
 */
void process_args(int *argc, char **argv[])
{
   int option;
   
   position=strdup("mouse"); //initialize
   while ((option=getopt(*argc,*argv,"p:c:skh"))!=-1)
     {
	switch(option)
	  {
	   case 'p':
	     position=strdup(optarg);
	     break;
	     
	   case 'c':
	     max_num_history=atoi(optarg);
	     break;
	     
	   case 's':
	     single=1;
	     break;
	     
	   case 'k':
	     keep=1;
	     break;
	     
	   case 'h':
	     show_help(*argv[0]);
	     break;
	  }//switch
     }//while
   
   /* since we always count from 0, subtract 1 from max_num_history.
    * this works cos even if we didnt specify -c on the command line,
    * the default was still set
    */
   max_num_history--;   
} //process_args

int main(int argc, char *argv[])
{
   GtkWidget *t_entry=NULL;
   GtkWidget *t_list=NULL;
   
   initialize();

   gtk_init(&argc, &argv);
   process_args(&argc, &argv);

   /* check if we need to exit. if -s is specified, we have to 
    * exit if a quickrun is already running
    */
   if (check_single() <= 0)
     {	
	printf("Another copy of quickrun is already running. Exiting now.\n");
	exit(1);
     }
      
   /* create main window */
   window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
      
   /* display the window where the user wants */
   if (!strcasecmp(position,"none"))
     gtk_window_set_position(GTK_WINDOW(window),GTK_WIN_POS_NONE);
   else
     if ((!strcasecmp(position,"center")) || (!strcasecmp(position,"centre")))
       gtk_window_set_position(GTK_WINDOW(window),GTK_WIN_POS_CENTER);
   else
     {
	if (!strcasecmp(position,"mouse"))
	  gtk_window_set_position(GTK_WINDOW(window),GTK_WIN_POS_MOUSE);
	else
	  {
	     printf("Specified window position not recognized. Legal values: \"none\" and \"center\"\n");
	     printf("Window position defaulting to \"mouse\"\n");
	     gtk_window_set_position(GTK_WINDOW(window),GTK_WIN_POS_MOUSE); //default
	  }
     }
   
   gtk_window_set_title(GTK_WINDOW(window),"Quick Run");
   gtk_signal_connect(GTK_OBJECT(window),"destroy",GTK_SIGNAL_FUNC(delete_window),NULL);   
   
   /* create vbox. this will contain the hboxes */
   mainvbox=gtk_vbox_new(FALSE,0);
   gtk_container_add(GTK_CONTAINER(window), mainvbox);
   
   /* create an hbox */
   mainhbox=gtk_hbox_new(FALSE,0);
   
   /* create and add stuff to hbox */
   /* label */
   label=gtk_label_new("Command to run: ");
   gtk_container_add(GTK_CONTAINER(mainhbox),label);
   gtk_widget_show(label);
   
   /* read history list */
   list=get_history_list(homedir);

   /* create combo box */
   combobox=gtk_combo_new();
   gtk_combo_set_use_arrows_always(GTK_COMBO(combobox), TRUE);
   gtk_combo_set_popdown_strings(GTK_COMBO(combobox), list);
   gtk_combo_set_value_in_list(GTK_COMBO(combobox), FALSE, TRUE);
   gtk_combo_disable_activate(GTK_COMBO(combobox));
   t_entry=GTK_COMBO(combobox)->entry; //point to textbox and list
   t_list=GTK_COMBO(combobox)->list;
   gtk_entry_set_text(GTK_ENTRY(t_entry),"");
   gtk_signal_connect(GTK_OBJECT(t_entry),"activate",
		      GTK_SIGNAL_FUNC(combobox_enter),NULL);
   gtk_signal_connect(GTK_OBJECT(t_entry),"key-press-event",
		      GTK_SIGNAL_FUNC(text_complete),NULL);
   gtk_container_add(GTK_CONTAINER(mainhbox),combobox);
   gtk_widget_show(combobox);
   
   /* add hbox to vbox */
   gtk_container_add(GTK_CONTAINER(mainvbox),mainhbox);
   gtk_widget_show(mainhbox);
   
   /* 
    * create and add a second hbox. this will only be filled with 
    * the list of completions if the user uses autocomplete
    */
   completionhbox=gtk_hbox_new(FALSE,0);

   /* add and show completionhbox */
   gtk_container_add(GTK_CONTAINER(mainvbox),completionhbox);
   gtk_widget_show(completionhbox);
   
   /* show vbox and window */
   gtk_widget_show(mainvbox);   
   gtk_widget_show(window);
   
   /* set focus to text box so that we can start typing in it immediately */
   gtk_widget_grab_focus(t_entry);
   
   /* set up keybindings for widgets */   
   /* set up accelerator for window */
   accel_group_window=gtk_accel_group_new();
   gtk_widget_add_accelerator(window, "destroy", accel_group_window,
			      GDK_q, GDK_MOD1_MASK, 0);
   gtk_window_add_accel_group(GTK_WINDOW(window), accel_group_window);
      
   gtk_main();  
  return 0;
}
